/* src/2ti86.c - Convert existing files to TI-86 Graph Link files
   Copyright (C) 2003, Dean Scarff.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

   Written 08 Mar 2003 by Dean Scarff <p00ya@ATusers.sf.DOTnet> */

/* Consult http://gladstone.uoregon.edu/~tsinger/link/TI-86/index.html for
   information on the TI-86 graph link format. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if HAVE_INTTYPES_H
# include <inttypes.h>
#else /* Not HAVE_INTTYPES_H */
# if HAVE_STDINT_H
#  include <stdint.h>
# endif
#endif /* Not HAVE_INTTYPES_H */
#include <stdio.h>
#include <errno.h>

#include <getopt.h>
#include <error.h>
#include "xmalloc.h"

/* Deal with variable argument obscurities. */
#if HAVE_VPRINTF || _LIBC
# if __STDC__
#  include <stdarg.h>
#  define VA_START(args, lastarg) va_start(args, lastarg)
# else
#  include <varargs.h>
#  define VA_START(args, lastarg) va_start(args)
# endif
#else
# define va_alist a1, a2, a3, a4, a5, a6, a7, a8
# define va_dcl char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
#endif /* HAVE_VPRINTF || _LIBC */

#include "titypes.h"

/* Use offsets for file positioning when available.  SEEK_T is the type used
   for seek and tell operations. */
#if HAVE_FSEEKO && HAVE_OFF_T
# define SEEK_T off_t
# define fseek fseeko
# define ftell ftello
#else
# define SEEK_T long int
#endif

/* Handle host-specific integer representations.  The TI-86 Graph Link format
   desires 2-byte, little-endian integers. */

#if HAVE_UINT16_T && !defined UINT16
# define UINT16 uint16_t
#else /* Not HAVE_UINT16_T */
# if SIZEOF_UNSIGNED_SHORT == 2
#  define UINT16 unsigned short;
# else
#  error "sanity failed (could not find unsigned 2-byte integer type)"
# endif
#endif /* Not HAVE_UINT16_T */

/* Swap the lower two bytes if the host is  big-endian, otherwise leave alone.
   This leaves an unsigned, little-endian 16-bit integer. */
#ifdef WORDS_BIGENDIAN
# define LITTLE16(n) ((UINT16) ((((n) & 0xff00U) >> 8) | (((n) & 0xffU) << 8);
#else
# define LITTLE16(n) (n)
#endif

/* getopt_long return codes. */
enum
{
  DUMMY_CODE = 129,
  BRIEF_CODE
};

/* The name the program was run with. */
char *program_name;

/* Output file. */
static char *oname = "STDOUT";
FILE *fout;

/* Conversion format. */
static char want_fmt[9] = "none";

/* Variable name.  TI-86 variables have a maximum length of nine characters. */
static char want_name[9] = "";

/* Volumes from --quiet --silent --verbose. */
static int want_volume = 1;

/* Input file. */
static char *iname = "STDIN";
FILE *fin;

static struct option const long_options[] =
{
  {"format", required_argument, 0, 'f'},
  {"name", required_argument, 0, 'n'},
  
  {"output", required_argument, 0, 'o'},

  {"quiet", no_argument, 0, 'q'},
  {"silent", no_argument, 0, 'q'},
  {"verbose", no_argument, 0, 'v'},
  {"brief", no_argument, 0, BRIEF_CODE},
  {"help", no_argument, 0, 'h'},
  {"usage", no_argument, 0, 'U'},
  {"version", no_argument, 0, 'V'},
  {NULL, 0, NULL, 0}
};

/* Help output. */
static void
help (int status)
{
  printf ("%s - \
Create TI-86 Graph Link files\n", program_name);
  printf ("\
Usage: %s [OPTION]... INPUT\n", program_name);
  printf ("\n\
Options:\n\
  -f, --format CONVERSION    convert using CONVERSION\n\
  -n, --name NAME            create variable with NAME\n\
\n\
  -o, --output FILE          place the output into FILE\n\
\n\
  -q, --quiet, --silent      inhibit output\n\
  -v, --verbose              increase output level\n\
      --brief                decrease output level\n\
  -h, --help                 display this help and exit\n\
  -U, --usage		     display usage summary and exit\n\
  -V, --version              print license and version then exit\n");

  /* List the available format conversions. */
  {  
    int i;

    /* String buffers. */
    char *name_buf, *desc_buf;
    
    char pad_buf[27];
    
    /* Stores the string length of name_buf. */
    int len;
        
    printf("\n\
Conversions:");
    for (i = 0; i < nformats; i++)
      {
	len = ti_format_desc (i, &name_buf, &desc_buf);
	printf ("\n\
  %s", name_buf);
	memset (pad_buf, ' ', 27 - len);
	printf ("%s%s", pad_buf, desc_buf);	
      }
  }
  
  printf ("\n\n\
%s Copyright (c) 2003 Dean Scarff.\n\
Please report bugs to <%s>\n", program_name, PACKAGE_BUGREPORT);
  exit (status);
}


/* Usage output. */
static void
usage (int status)
{
  printf ("%s - Create TI-86 Graph Link files", program_name);
  printf ("\n\
Usage: %s -[fn ARG] [-o FILE] -[qvhUV]... INPUT\n", program_name);
  exit (status);
}

/* Version output. */
static void
version (int status)
{
  printf ("%s - part of %s (%s)", program_name, PACKAGE_STRING, TARGET);
  printf ("\n\
Copyright (c) 2003 Dean Scarff.\n\
%s comes with ABSOLUTELY NO WARRANTY.", PACKAGE_NAME);
  printf ("\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n");
  exit (status);
}

/* Information output.  If VOLUME is <= WANT_VOLUME then print MESSAGE, which
   is a printf-style template string with optional arguments. */

static void
#if defined VA_START && __STDC__
info (int volume, const char *message, ...)
#else /* Not VA_START && __STDC__ */
info (volume, message, va_alist)
     int volume;
     char *message;
     va_dcl
#endif /* Not VA_START && __STDC__ */
{
  if (volume > want_volume)
    return;

  fflush (stdout);
  printf ("%s: ", program_name);
  
#ifdef VA_START
  va_list args;
  VA_START (args, message);

# if HAVE_VPRINTF
  vprintf (message, args);
# else
  _doprnt (message, args, stdout); 
# endif
  va_end (args);  
#else /* Not VA_START */
  printf (message, a1, a2, a3, a4, a5, a6, a7, a8);
#endif /* Not VA_START */

  putc ('\n', stdout);

  fflush (stdout);  
}

/* Option and argument parsing. */

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument. */
static void
decode_switches (int argc, char **argv)
{
  int c;

  while ((c = getopt_long (argc, argv,
			   "f:"	/* format */
			   "n:" /* name */

			   "o:"	/* output */

			   "q"	/* quiet or silent */
			   "v"	/* verbose */
			   "h"	/* help */
			   "U"	/* usage */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'f':
	  if (sscanf (optarg, "%8s", want_fmt) != 1)
	    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
			   "error parsing format \"%s\"", optarg);
	  info (4, "Format set to \"%s\" from options", want_fmt);
	  
	  break;

	case 'n':
	  if (sscanf (optarg, "%8s", want_name) != 1)
	    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
			   "error parsing variable name \"%s\"", optarg);
	  info (4, "Variable name set to \"%s\" from options", want_name);
	  break;

	case 'o':
	  oname = xstrdup (optarg);
	  info (4, "Output file set to \"%s\" from options", oname);
	  break;

	case 'q':
	  want_volume = 0;
	  break;

	case 'v':
	  want_volume++;
	  break;

	case BRIEF_CODE:
	  want_volume--;
	  break;

	case 'V':
	  version (0);

	case 'U':
	  usage (0);

	case 'h':
	  help (0);

	default:
	  help (EXIT_FAILURE);
	}
    }
}

static void
decode_args (int argc, char **argv)
{
  /* TODO: This will change to allow more than one INPUT file. */
  if ((argc - optind) > 1)
    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		   "Too many INPUT arguments.");
  
  iname = xstrdup (argv[optind]);
  info (4, "Input file set to \"%s\" from argument", iname);
}

/* Data checksum. */
static UINT16 chk = 0;

/* Data length. */
static UINT16 data_len = 0;

/* Write data from BUF of length LEN to file FP,  keeping track of the
   checksum. */
void
datawrite (const void *buf, int len)
{
  int i;

  fwrite(buf, len, 1, fout);
  if (ferror (fout))
    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		   "error writing data to \"%s\"", oname);
  for(i = 0; i < len; i++)
    chk += ((unsigned char *)buf)[i];
  data_len += len;
}


int main(int argc, char* argv[])
{
  program_name = strrchr (argv[0], '/');
  program_name = (program_name) ? program_name + 1 : argv[0];

  fout = stdout;
  fin = stdin;
    
  decode_switches (argc, argv);
  decode_args (argc, argv);

  /* Open input and output streams. */
  info (3, "Opening file \"%s\" for output", oname);
  if (strcmp (oname, "STDOUT"))
    fout = fopen (oname, "wb");
  if (!fout)
    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		   "cannot open %s for writing", oname);

  info (3, "Opening file \"%s\" for input", iname);
  
  if (strcmp (iname,"STDIN"))
    fin = fopen (iname, "rb");
  if (!fin)
    error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		   "cannot open %s for reading", iname);

  info (4, "Writing the static file header.");
  
  /* File header. */
  fwrite("**TI86**\x1a\x0a\x00", 11, 1, fout);
  
  info (4, "Writing the file comment.");  
  /* Write the comment. */
  {
    char buf[42];
    int offset = 17;
    /* Omit the path up to and including the '/' separator. */
    char *input_name;
			
    input_name = strrchr (iname, '/');
    input_name = (input_name) ? input_name + 1 : iname;
   
    sprintf (buf, "Produced by 2ti86");
    if (strlen (input_name) < 20)
      {
	sprintf (buf + offset, " from %s", input_name);
	offset += 6;
	offset += strlen (iname);
      }
    
    /* Pad comment right with null. */
    memset (buf + offset, '\x00', 42 - offset);
    fwrite (buf, 1, 42, fout);
  }
 
  /* NOTE: Come back and write the data length. */
  fseek (fout, 2, SEEK_CUR);

  info (4, "Writing the file data section (variables).");
  
  /* WISHLIST: More than one variable here. */
  /* This block writes the data section of the file.  This can comprise of a
     number of variable entries, or (as supported at the moment) a single
     variable entry. */ 
  {
    /* One character general purpose buffer. */
    char cbuf;
    /* Size of the variable data. */
    UINT16 vardata_len;

    /* Start of the variable within the file. */
    SEEK_T var_start = ftell (fout);

    /* Format index. */
    size_t fmt_ind = ti_format_ind (want_fmt);
    info (4, "Format has index %u", fmt_ind);
                
    datawrite ("\x0c\x00", 2);

    /* NOTE: Come back to write the variable data length. */
    fseek (fout, 2, SEEK_CUR);
        
    /* Variable type identifier. */
    cbuf = ti_type_id (fmt_ind);
    datawrite (&cbuf, 1);
    info (4, "Added variable type identifier 0x%x", cbuf);
    
    /* Name. */
    {
      /* Length of the (potential) name. */
      unsigned char name_size;
           
      /* Generate a variable name, if one is not already present. */
      if (!strcmp (want_name, ""))
	{
	  /* Extract the filename from INAME without the path or extension.  If
	     this is more than eight characters it will be truncated. */
	  char *name_start;
	  char *name_end;
			
	  name_start = strrchr (iname, '/');
	  name_start = (name_start) ? name_start + 1 : iname;

	  name_end = strrchr (name_start, '.');
	  name_size = name_end - name_start;
	  	
	  if (name_end && name_size <= 8)
	    memcpy (want_name, name_start, name_size);
	  else
	    {
	      name_size = 8;
	      memcpy (want_name, name_start, 8);
	    }
	}
      else
	name_size = (unsigned char) strlen (want_name);

      /* Pad want_name right with null. */
      memset (want_name + name_size, '\x00', 8 - name_size);
            
      info (4, "Writing variable name \"%-8s\"", want_name);
      /* Write the name length. */
      datawrite (&name_size, 1);
      /* Write the name. */
      datawrite (want_name, 8);
    }

    /* NOTE: Come back to write the variable data length. */
    fseek (fout, 2, SEEK_CUR);

    info (2, "writing from \"%s\" to \"%s\" with format conversion \"%s\"",
	    iname, oname, want_fmt);
    /* The following block converts the entirity of FIN according to the
       conversion format.  First of all FIN is read to a buffer.  */
    {     
      size_t in_size;
      char *data_buf;
      char c;      
      FILE *tmp;    
      
      /* Find the size of fin.  There is some suggestion that ftell will not
         return the number of bytes on some systems, and may instead return
         some sort of internal encoding for a record and an offset.  I have
	 not observed this behaviour. */
      fseek (fin, 0, SEEK_END);
      in_size = (size_t) ftell (fin);
      rewind (fin);

      data_buf = (char *) xmalloc (in_size);
      fread(data_buf, in_size, 1, fin);
      if (ferror(fin))
	error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		       "error reading from input \"%s\"", iname);      
      fclose(fin);
      
      /* Create a temporary file to receive variable data. */
      tmp = tmpfile();
      if (tmp == NULL)
	error_at_line (EXIT_FAILURE, errno, __FILE__, __LINE__,
		       "could not create temporary file");      
	  
      ti_convert (fmt_ind, tmp, data_buf, in_size);
      
      rewind (tmp);

      /* Save the offset of the variable data within the data section. */
      vardata_len = data_len;
      
      while ((c = fgetc (tmp)) != EOF)
	datawrite (&c, 1);
      fclose (tmp);

      free (data_buf);
      
    }

    /* Go back and record the variable data lengths. */
    {
      /* The end of the variable within the file. */     
      vardata_len = data_len - vardata_len;
      info (4, "Writing variable length %u", vardata_len);
      vardata_len = LITTLE16(vardata_len);      
      
      fseek (fout, var_start + 2, SEEK_SET);
      datawrite (&vardata_len, 2);
      fseek (fout, var_start + 14, SEEK_SET);
      datawrite (&vardata_len, 2);

      fseek (fout, 0, SEEK_END);
    }
  }
  
  
  {
    /* Go back and write the data length. */
    info (4, "writing data length %u", data_len);    

    data_len = LITTLE16(data_len);
    fseek (fout, 53, SEEK_SET);
    fwrite (&data_len, 2, 1, fout);

    /* Go to the end and write the data checksum. */
    chk = LITTLE16(chk);
    fseek (fout, 0, SEEK_END);
    fwrite (&chk, 2, 1, fout);
  }
  fclose (fout);
  fclose (fin);
  
  info (2, "Graph Link file \"%s\" successfully created!", oname);
  
  return EXIT_SUCCESS;
}

