/* src/titypes.c - conversion tables for TI-86 formats.
   Copyright (C) 2003, Dean Scarff.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

   Written 8 Mar 2003 by Dean Scarff <p00ya@ATusers.sf.DOTnet> */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#ifdef	HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include <errno.h>
#if HAVE_INTTYPES_H
# include <inttypes.h>
#else /* Not HAVE_INTTYPES_H */
# if HAVE_STDINT_H
#  include <stdint.h>
# endif
#endif /* Not HAVE_INTTYPES_H */
#include <stdio.h>

#include <error.h>
#include "xmalloc.h"

#include "titypes.h"

/* Use offsets for file positioning when available.  SEEK_T is the type used
   for seek and tell operations. */
#if HAVE_FSEEKO && HAVE_OFF_T
# define SEEK_T off_t
# define fseek fseeko
# define ftell ftello
#else
# define SEEK_T long int
#endif

/* Handle host-specific integer representations.  The TI-86 Graph Link format
   desires 2-byte, little-endian integers. */

#if HAVE_UINT16_T && !defined UINT16
# define UINT16 uint16_t
#else /* Not HAVE_UINT16_T */
# if SIZEOF_UNSIGNED_SHORT == 2
#  define UINT16 unsigned short;
# else /* Not HAVE_UNSIGNED_SHORT */
#  error "sanity failed (could not find unsigned 2-byte integer type)"
# endif /* Not HAVE_UNSIGNED_SHORT */
#endif /* Not HAVE_UINT16_T */

/* Swap the lower two bytes if the host is  big-endian, otherwise leave alone.
   This leaves an unsigned, little-endian 16-bit integer. */
#ifdef WORDS_BIGENDIAN
# define LITTLE16(n) ((UINT16) ((((n) & 0xff00U) >> 8) | (((n) & 0xffU) << 8);
#else
# define LITTLE16(n) (n)
#endif

/* Type conversion functions. */
static void
ti_convert_s_86 (size_t fmt_index, FILE *streamout,
		 const char *in_buf, size_t len);

/* Types table.  Order by type_id. */
const size_t ntypes = 2;
static struct ti_type const types_86[] = 
{
  /* Strn vars. */
  {'\x0c', &ti_convert_s_86, 's'},
  /* Prgm vars. */
  {'\x12', &ti_convert_s_86, 'p'}
};


/* Formats table.  Order by desc. */
const size_t nformats = 2;
static struct ti_format const formats_86[] =
{
  {
    "none", "Raw program, no substitutions (default)", '\x12', 0, NULL,
    {{'\0', NULL}}
    
  },
  {
    "text",
    "Text header and ASCII/Win-XASCII substitutions",
    '\x12',
    1,
    "\x00",
    {
      /* Line control.  Drop carriage-returns. */
      {'\x0a',"\xd6"},{'\x0d',""},
      /* Soft-hypen -> Negator. */
      {'\xad',"\x1a"},
      /* B-row (superscript, miscellaneous. */
      {'\xb2',"\x12"},{'\xb3',"\xd5"},{'\xba',"\x14"},{'\xbf',"\xb9"},
      /* Upper-case accented A. */
      {'\xc0',"\x8b"},{'\xc1',"\x8a"},{'\xc2',"\x8c"},{'\xc3',"\x8d"},
      /* Upper-case C-dilla. */
      {'\xc7',"\xb2"},	
      /* Upper-case accented E. */
      {'\xc8',"\x94"},{'\xc9',"\x93"},{'\xca',"\x95"},{'\xcb',"\x96"},
      /* Upper-case accented I. */
      {'\xcc',"\x9b"},{'\xcd',"\x9a"},{'\xce',"\x9c"},{'\xcf',"\x9d"},
      /* Upper-case accented O. */
      {'\xd2',"\xa4"},{'\xd3',"\xa3"},{'\xd4',"\xa5"},{'\xd2',"\xa6"},
      /* Upper-case accented U. */
      {'\xd9',"\xab"},{'\xda',"\xaa"},{'\xdb',"\xac"},{'\xdc',"\xad"},
      /* Beta. */
      {'\xdf',"\xbc"},
      /* Lower-case accented a. */
      {'\xe0',"\x8f"},{'\xe1',"\x8e"},{'\xe2',"\x90"},{'\xe4',"\x91"},
      /* Lower-case accented e. */
      {'\xe8',"\x97"},{'\xe9',"\x96"},{'\xea',"\x98"},{'\xeb',"\x99"},
      /* Lower-case accented i. */
      {'\xec',"\x9f"},{'\xed',"\x9e"},{'\xee',"\xa0"},{'\xef',"\xa1"},
      /* Lower-case accented o. */
      {'\xf2',"\xa7"},{'\xf3',"\xa6"},{'\xf4',"\xa8"},{'\xf5',"\xa9"},
      /* Lower-case accented u. */
      {'\xf9',"\xaf"},{'\xfa',"\xae"},{'\xfb',"\xb0"},{'\xfc',"\xb1"},

      /* Terminator. */
      {'\0', NULL}
    }
    
  }
};

/* The following 'comparison' functions should only be called with a key then
   the appopriate structure as an argument.  This means they cannot be used
   in a qsort() call. */

/* Compares character A to the from value in a subst_pair. */
static int
compar_substs (const void *a, const void *b)
{
  const unsigned char _a = *(const char *) a;
  const unsigned char _b = (unsigned char) ((struct subst_pair *) b)->from;

  return _a - _b;
}

/* Compares the string A with a ti_format's name B. */
static int
compar_namedformats (const void *a, const void *b)
{
  const char *_a = *(const char **) a;
  const char *_b = (const char *) ((const struct ti_format *) b)->name;
  
  return strcasecmp (_a, _b);
}

/* Compares the character A with a ti_type's type_id B. */
static int
compar_types (const void *a, const void *b)
{
  const unsigned char _a = *(const char *) a;
  const unsigned char _b = (unsigned char) ((struct ti_type *) b)->type_id;
  
  return _a - _b;
}

/* The pointer value of TO is set to the replacement string using the
   substitutions defined by the format indexed by INDEX, for the character
   FROM.  The length of the replacment string is returned.
   NOTE: memory will be allocated at the address referenced by *to.  No checks
   are made to see whether TO should be freed first. */
static size_t
ti_replace (size_t index, char **to, const char from)
{
  struct subst_pair *subst;
  
  /* The number of non-blank substitution pairs for the format. */
  size_t nsubs = 0;
  char *subst_to;

  if (index >= nformats)
    error_at_line (0, EINVAL, __FILE__, __LINE__,
		   "index %u exceeds the number of available formats (%u)",
		   index, nformats);

  /* Find NSUBS. */
  for (int i = 0; formats_86[index].subs[i].from != '\0'; i++)
    nsubs++;
  
  subst = (struct subst_pair *) bsearch (&from, formats_86[index].subs, nsubs,
					 sizeof (struct subst_pair), 
					 compar_substs);

  if (subst == NULL)
    {     
      *to = xmalloc (2);
      (*to)[0] = from;
      (*to)[1] = '\0';
      return 1;
    }      
  
  subst_to = (char *) subst->to;
  
  *to = xstrdup (subst_to);
  return strlen (*to);
}

/* Places the variable data header to *HEADER for the formats_86 element
   with the given INDEx.  No memory is allocated.  Returns the string length
   of HEADER. */
static size_t
ti_tag (size_t index, char **tag)
{
  if (index >= nformats)
    error_at_line (0, EINVAL, __FILE__, __LINE__,
		   "index %u exceeds the number of available formats (%u)",
		   index, nformats);

  *tag = (char *) formats_86[index].tag;

  return formats_86[index].tag_len;
}

/* Return the index of a format with its name equal to NAME.  Use this to find
   the index argument used in later functions. */
size_t
ti_format_ind (const char *name)
{
  struct ti_format *fmt;
  
  fmt = (struct ti_format *) bsearch (&name, formats_86, nformats,
				      sizeof (struct ti_format),
				      compar_namedformats);
  if (fmt == NULL)
    error_at_line (EXIT_FAILURE, EINVAL, __FILE__, __LINE__,
		   "could not find format \"%s\"", name);
  /*  return (size_t) (fmt - formats_86) / sizeof (struct ti_format); */
  /* FIXME: Return value was AFU with the above... */
  return (size_t) (fmt - formats_86);
  
}

/* Return the variable type ID for the format of index INDEX. */
char
ti_type_id (size_t index)
{
  struct ti_type *typ;
  if (index >= nformats)
    error_at_line (0, EINVAL, __FILE__, __LINE__,
		   "index %u exceeds the number of available formats (%u)",
		   index, nformats); 
  
  typ = (struct ti_type *) bsearch (&(formats_86[index].type_id),
				    types_86, ntypes, sizeof (struct ti_type),
				    compar_types);
  if (typ == NULL)
    error_at_line (EXIT_FAILURE, EINVAL, __FILE__, __LINE__,
		   "\n\
could not find type with ID 0x%x referenced by format of index %u",
		   formats_86[index].type_id, index);
  
  return typ->type_id;
}

/* Places the description to *DESC and the name to *NAME for the formats_86
   element with the given INDEX.  No memory is allocated.
   Returns the string length of NAME. */
size_t
ti_format_desc (size_t index, char **name, char **desc)
{
  if (index >= nformats)
    error_at_line (0, EINVAL, __FILE__, __LINE__,
		   "index %u exceeds the number of available formats (%u)",
		   index, nformats);

  *name = (char *) formats_86[index].name;
  *desc = (char *) formats_86[index].description;
  return strlen (*name);
}

/* Wrapper conversion function that will call appropriate conversion function
   depending on the type defined in format. */
void
ti_convert (size_t index, FILE *streamout, const char *in_buf, size_t len)
{
  struct ti_type *typ;
  if (index >= nformats)
    error_at_line (0, EINVAL, __FILE__, __LINE__,
		   "index %u exceeds the number of available formats (%u)",
		   index, nformats); 
  
  typ = (struct ti_type *) bsearch (&(formats_86[index].type_id),
				    types_86, ntypes, sizeof (struct ti_type),
				    compar_types);
  if (typ == NULL)
    error_at_line (EXIT_FAILURE, EINVAL, __FILE__, __LINE__,
	   "could not find type 0x%x referenced by format of index %u",
	   formats_86[index].type_id, index);
  
  (*(typ->convert)) (index, streamout, in_buf, len);

  return;
}

/* Conversion for TI-86 strings and other string-derived types.  Using the
   format indexed by FMT_INDEX, writes converted data from IN_BUF back along
   STREAMOUT.  STREAMOUT must support writes and seeks, and the caller is
   responsible for closing it. */
static void
ti_convert_s_86 (size_t fmt_index, FILE *streamout,
		 const char *in_buf, size_t len)
{
  /* The size of the 86s-string. */
  size_t s_86_z = 0;

  /* The offset of the variable data within STREAMOUT. */
  SEEK_T vardata_start = ftell (streamout);
           
  /* NOTE: Come back to write the 2-byte length header. */
  fseek (streamout, 2, SEEK_CUR);
  
  /* Write the variable data tag. */
  {
    char *sbuf;
    size_t slen = ti_tag (fmt_index, &sbuf);       
    fwrite (sbuf, slen, 1, streamout);

    s_86_z += slen;
  }
        
  /* Read from IN_BUF and write char-wise. */
  for (size_t i = 0; i < len; i++)
    {
      /* String buffer to receive substitutions. */
      char *sbuf;
      size_t slen;
      
      slen = ti_replace (fmt_index, &sbuf, in_buf[i]);
	
      fwrite (sbuf, slen, 1, streamout);
      free (sbuf);

      s_86_z += slen;
    }
  {   
    UINT16 s_86_len = LITTLE16(s_86_z);
      
    fseek (streamout, vardata_start, SEEK_SET);
    fwrite (&s_86_len, 2, 1, streamout);
  }
  
  fseek (streamout, 0, SEEK_END);

  return;
}
